@@ -1,14 +1,31 @@
-import socket
-import asyncio
-from typing import Union
-import os
-import json
 from .helper import http_status_codes, default_headers
 from .utlies import printc, Colors
+from typing import Union
+import asyncio
+import socket
+import uvloop
+import json
+import os
 import re
 
+class TemplateFolderNotFound(Exception):
+    pass
+
+class TemplateNotFound(Exception):
+    pass
+
+class MultiPart:
+    def __init__(self) -> None:
+        """
+        Layout is just what data is representing, there indexes are 
+        the same just one tell you info about the other
+        """
+        self.layout = []
+        self.data = []
+
 class Connection:
     def __init__(self, request: dict, args: dict = {}) -> None:
+        __slots__ = ('request', '_response', 'args')
         self.request = request
         self._response = ['HTTP/1.1 {status_code} {status}\r\n', '', b'\r\n', b'']
         self.args = args
@@ -35,13 +52,58 @@ class Connection:
 
 class Lamp:
     def __init__(self) -> None:
+        __slots__ = (
+            'routes', 'error_handlers',
+            'templateDir', 'regex',
+            'debug', 'uv'
+        )
         self.routes = {}
         self.error_handlers = {}
+        self.templateDir = './templates'
+        self.regex = {
+            'htmlVars': re.compile(r'\[\[(.*)\]\]'),
+            'types': re.compile(r'^<(.*)>$')
+        }
+        self.debug = False
+        self.uv = False
+         
+    def renderTemplate(self, file, **kwargs):
+        """
+        There are problably better implementations of 
+        this type of stuff, but its always good to give
+        it a try!
+        """
+        if not os.path.exists(self.templateDir):
+            raise TemplateFolderNotFound()
+        if not os.path.exists(f"{self.templateDir}/{file}"):
+            raise TemplateNotFound()
+        
+        with open(f"{self.templateDir}/{file}", 'r') as f:
+            html = f.read().splitlines() # get content of html file
+            h = '\n'.join(html) # make a copy so py doesn't reference the html variable
+
+        for index, line in enumerate(h.splitlines()): # read every line to see if we need to parse it
+            
+            if (x := self.regex['htmlVars'].findall(line)):
+                for v in x:
+                    if (vv := v.strip()) in kwargs:
+                        line = line.replace(f"[[{v}]]", kwargs[vv])
+                html[index] = line
+
+            if 'link' in line and (x := self.regex['types'].findall(line)):
+                l = x[0].split()[1:]
+                link = {k: v[1:][:-1] for k, v in [z.split('=', 1) for z in l]}
+                if False not in [x in tuple(link) for x in ('type', 'rel', 'href')]:
+                    if (link['type'], link['rel']) == ('css', 'stylesheet'):
+                        with open(f"{self.templateDir}/{link.get('href')}", 'r') as css:
+                            html[index] = '<style>' + css.read() + '</style>'
+        
+        return '\n'.join(html)
 
     def route(self, route: str, domain: Union[str, bool] = None, method: list = []):
         def inner(func):
             
-            key = json.dumps({'route': route, 'domain': domain, 'method': method})
+            key = json.dumps({'route': str(route), 'domain': domain, 'method': method})
 
             self.routes[key] = func
             
@@ -57,20 +119,28 @@ class Lamp:
         return inner
     
     def parse_multi(self, boundary: str, body: bytes) -> dict:
-        disposition = {}
+        multi = MultiPart()
         body = [x for x in body.split(boundary.encode()) if x and x != b'--\r\n']
+
         for arg in body:
+            data = []
+            layout = []
+
             arg, content = arg.split(b'\r\n\r\n', 1)
             arg = arg.decode().split('; ')[1:]
-            index = 0
+
             for x in arg:
                 x = x.split('=', 1)
                 x[1] = x[1].strip('"')
-                arg[index] = {k: v for k, v in [x]}
-                index += 1
-            arg.append(content)
-            disposition[len(disposition) + 1] = arg
-        return disposition
+                layout.append(x[0])
+                data.append(x[1])
+
+            layout.append('content')
+            data.append(content)
+            multi.layout.append(layout)
+            multi.data.append(data)
+
+        return multi
     
     async def parse(self, req: bytes) -> dict:
         r = {}
@@ -102,20 +172,22 @@ class Lamp:
         _req = await loop.sock_recv(client, 1024)
         if not _req:
             return
-        req = await self.parse(_req)
-        if 'Content-Length' in req and int(req['Content-Length']) > 1024:
-            _req += await loop.sock_recv(client, 1000000)
-            req = await self.parse(_req)
-        
+        req = await self.parse(_req)        
+        if 'Content-Length' in req:
+            c = int(req['Content-Length'])
+            if c > 1024:
+                _req += await loop.sock_recv(client, c)
+                req = await self.parse(_req)
+
         for key in self.routes:
-            k = json.loads(key) # brained Lawl
+            k = json.loads(key)
             r = []
             args = {}
-            # checks
 
-            if k['route'][0] == '^':
-                k['route'] = re.compile(k['route'])
-                x = k['route'].match(req['path'])
+            if k['route'].startswith('re.compile'):
+                x = {}
+                exec('import re ; x = ' + k['route'], x)
+                x = x['x'].match(req['path'])
                 if x:
                     args = x.groupdict()
                     for K, V in args.items():
@@ -130,13 +202,24 @@ class Lamp:
                     r.append(True)
                 else:
                     r.append(False)
+                
+            if k['domain'] and 'Host' in req:
+                if k['domain'].startswith('re.compile'):
+                    exec('import re ; x = ' + k['domain'], (x := {}))
+                    x = x.match(req['path'])
+                    if x:
+                        args = x.groupdict()
+                        for K, V in args.items():
+                            if not K or not V:
+                                r.append(False)
+                            else:
+                                r.append(True)
 
-            if k['domain']:
-                if 'Host' in req and req['Host'] == k['domain']:
+                elif req['Host'] == k['domain']:
                     r.append(True)
                 else:
                     r.append(False)
-            
+
             if k['method']:
                 if req['method'] in k['method']:
                     r.append(True)
@@ -146,13 +229,15 @@ class Lamp:
             if False in r or True not in r:
                 continue
             else:
-                printc(f"Request for: {req['Host']} | {req['path']} | {req['method']} | {req['http_vers']}", Colors.Green)
+                if self.debug:
+                    printc(f"{req['Host']} | {req['path']} | {req['method']} | {req['http_vers']}", Colors.Green)
                 await loop.sock_sendall(client,
                     await self.routes[key](Connection(req, args))
                 )
                 return
 
-        printc(f"Unhandled Request for: {req['Host']} | {req['path']} | {req['method']} | {req['http_vers']}", Colors.Red)
+        if self.debug:
+            printc(f"{req['Host']} | {req['path']} | {req['method']} | {req['http_vers']}", Colors.Red)
         if 404 in self.error_handlers:
             er = self.error_handlers[404](Connection(req))
         else:
@@ -161,7 +246,7 @@ class Lamp:
         await loop.sock_sendall(client, er)
         return
 
-    def run(self, socket_type: Union[str, tuple]) -> None: # I need to find a better word for this
+    def run(self, socket_type: Union[str, tuple], **kwargs) -> None:
         """
         socket_type: str or tuple
             You can decide wheather you want your web server to be on a
@@ -169,7 +254,20 @@ class Lamp:
             Examples:
                 run(socket_type = '/tmp/cover.sock') <- unix_socket
                 run(socket_type = ("127.0.0.1", 5000)) <- ip and port
+        
+        debug: bool = False
+            All this does for right now is print on the console
+            all the request that are coming in, even the bad ones
+        
+        uvloop: bool = False
+            Free speed pretty much.
         """
+        self.debug = kwargs.get('debug', False)
+        self.uv = kwargs.get('uvloop', False)
+
+        if self.debug:
+            printc('Debug Mode!', Colors.Green)
+
         async def _run():
             
             if not isinstance(socket_type, (str, tuple)):
@@ -183,17 +281,20 @@ class Lamp:
                 s = socket.AF_INET
 
             with socket.socket(s) as sock:
+                if self.uv:
+                    printc('Using uvloop!', Colors.Green)
+                    uvloop.install()    
                 loop = asyncio.get_event_loop()
                 sock.bind(socket_type)
-                if isinstance(socket_type, str): os.chmod(socket_type, 0o777)
+                if isinstance(socket_type, str): 
+                    os.chmod(socket_type, 0o777)
                 sock.listen(5)
                 sock.setblocking(False)
                 
-                print(f"Server is up and running on: {socket_type}") # unpack tuple
+                printc(f"Server is up and running on: {socket_type}", Colors.Green)
 
                 while True:
                     client, addr = await loop.sock_accept(sock)
-                    # await self.handler(client)
                     loop.create_task(self.handler(client, loop))
         
         asyncio.run(_run())
\ No newline at end of file
